from seeq import spy
import pandas as pd

from seeq.spy.assets import Asset

pd.set_option('display.max_colwidth', -1)

spy.login(username="nicholas.gigliotti@seeq.com", password="Seeq2019!", url="https://birlacarbon.seeq.site",
          auth_provider="Seeq")


class Site(Asset):
    @Asset.Component()
    def Units(self, metadata):
        unit_names = metadata['Unit'].dropna().drop_duplicates().tolist()
        unit_definitions = list()
        for unit_name in unit_names:
            unit_definition = Unit({
                'Name': unit_name,
            }, parent=self).build(metadata[metadata['Unit'] == unit_name])
            unit_definitions.extend(unit_definition)
        return unit_definitions


class Unit(Asset):
    @Asset.Component()
    def Reactors(self, metadata):
        reactor_names = metadata['Reactor'].dropna().drop_duplicates().tolist()
        reactor_definitions = list()
        for reactor_name in reactor_names:
            reactor_definition = Reactor({
                'Name': reactor_name,
            }, parent=self).build(metadata[metadata['Reactor'] == reactor_name])
            reactor_definitions.extend(reactor_definition)
        return reactor_definitions


class Reactor(Asset):
    @Asset.Attribute()
    def Make_Oil_Flow_Rate(self, metadata):
        return metadata[metadata['Description'].str.contains('MAKE OIL|MAKEOIL')]

    @Asset.Attribute()
    def Blast_Air_Flow_Rate(self, metadata):
        metadata1 = metadata[metadata['Description'].str.contains('BLAST AIR')]
        metadata2 = metadata1[~metadata1['Description'].str.contains('VENTURI')]

        if len(metadata2) != 1:
            return None

        if metadata2.iloc[0]['Value Unit Of Measure'] != 'ft³/h':
            return {
                'Name': 'Blast Air Flow Rate',
                'Type': 'Signal',
                'Formula': "($tag * 1000).setUnits('ft³/h')",
                'Formula Parameters': {
                    '$tag': metadata2
                },
            }
        else:
            return metadata2

    @Asset.Attribute()
    def Blast_Gas_Flow_Rate(self, metadata):
        metadata1 = metadata[metadata['Description'].str.contains('BLAST GAS')]
        return metadata1[~metadata1['Description'].str.contains('AUXILIARY')]

    @Asset.Attribute()
    def Oil_Status(self, metadata):
        metadata2 = metadata[metadata['Name'].str.contains('AN')]
        if len(metadata2) != 1:
            return None

        return {
            'Type': 'Signal',
            'Formula': "$tag.validValues().toStep().setmaxInterpolation(30 days)",
            'Formula Parameters': {
                '$tag': metadata2
            },
        }


unit1_vars = spy.search({'Name': '/^FIC1[1-4]+0[2-4]+\.PV.*/', 'Type': 'StoredSignal'})
oil_status_1 = spy.search({'Name': '/^XS1[1-5]+02C_AN$/', 'Description': '/OIL/'})
unit1_df = unit1_vars.append(oil_status_1)
other_units = spy.search({'Name': '/^FIC[2-5]+10[1-4]+.\.PV./'})
all_units = unit1_df.append(other_units)
all_units['Reactor'] = 'RX' + all_units['Name'].str.extract(pat='\w{2,3}(.).*')
all_units['Unit'] = 'Unit ' + all_units['Name'].str.extract(pat='\w{2,3}.(.).*')
all_units['Site'] = 'North Bend'
all_units['Instance Path'] = 'Nick Dev'
all_units['Instance Asset'] = 'North Bend'
instance_df = spy.assets.instantiate(Site, all_units)
print(instance_df)
push_results_df = spy.push(metadata=instance_df, errors='catalog')
push_results_df



import re

import pandas as pd
import numpy as np

from seeq.sdk import *

from .. import _common
from .. import _login


class Workbook:
    pass


class Analysis(Workbook):
    pass


def get_folder(path=None, content_filter='owner', recursive=False):
    return _get_folder(path, content_filter, recursive)


def _get_folder(path_filter, content_filter, recursive, parent_id=None, parent_path=None):
    folders_api = FoldersApi(_login.client)
    results_df = pd.DataFrame()

    path_filter_parts = None
    if path_filter is not None:
        path_filter_parts = re.split(r'\s*>>\s*', path_filter.strip())

    if parent_id is not None:
        folder_output_list = folders_api.get_folders(filter=content_filter,
                                                     folder_id=parent_id,
                                                     limit=10000)  # type: FolderOutputListV1
    else:
        folder_output_list = folders_api.get_folders(filter=content_filter,
                                                     limit=10000)  # type: FolderOutputListV1

    for content in folder_output_list.content:  # type: FolderOutputV1
        if path_filter_parts and content.name != path_filter_parts[0]:
            continue

        if not path_filter_parts:
            results_df = results_df.append({
                'ID': content.id,
                'Type': content.type,
                'Path': parent_path if parent_path else np.nan,
                'Name': content.name,
                'Description': content.type,
                'Access Level': content.access_level
            }, ignore_index=True)

        if content.type == 'Folder' and recursive:
            child_path_filter = None
            if path_filter_parts and len(path_filter_parts) > 1:
                child_path_filter = ' >> '.join(path_filter_parts[1:])

            if parent_path is None:
                new_parent_path = content.name
            else:
                new_parent_path = parent_path + ' >> ' + content.name

            child_results_df = _get_folder(child_path_filter, content_filter, recursive, content.id, new_parent_path)

            results_df = results_df.append(child_results_df,
                                           ignore_index=True)

        return results_df


def _put_analysis():
    pass



import pandas as pd

from .. import _common


class _AssetBase:

    def __init__(self, definition=None, parent=None):
        self.asset_definition = dict()

        if isinstance(definition, _AssetBase):
            self.asset_definition = definition.asset_definition
        elif isinstance(definition, pd.DataFrame):
            if len(definition) != 1:
                raise RuntimeError('DataFrame must be exactly one row')
            self.asset_definition = definition.iloc[0].to_dict()
        elif isinstance(definition, pd.Series):
            self.asset_definition = definition.to_dict()
        elif definition is not None:
            self.asset_definition = definition

        self.asset_definition['Type'] = 'Asset'
        if 'Name' in self.asset_definition:
            self.asset_definition['Asset'] = self.asset_definition['Name']

        self.parent_definition = parent.asset_definition if isinstance(parent, _AssetBase) else parent

        if self.parent_definition is not None:
            self.asset_definition['Path'] = self.parent_definition['Path'] + ' >> ' + self.parent_definition['Name']

    def build(self, metadata):
        definitions = list()
        object_methods = [getattr(self, method_name) for method_name in dir(self)
                          if callable(getattr(self, method_name))]

        for func in object_methods:
            if not hasattr(func, 'spy_model'):
                continue

            attribute = func(metadata)

            if attribute is None:
                continue

            if isinstance(attribute, list):
                definitions.extend(attribute)
            else:
                definitions.append(attribute)

        return definitions


class Mixin(_AssetBase):

    def __init__(self, definition):
        super().__init__(definition)

    def build(self, metadata):
        definitions = super().build(metadata)
        return definitions


class Asset(_AssetBase):

    def __init__(self, definition=None, parent=None):
        super().__init__(definition, parent)
        self.asset_definition['Template'] = self.__class__.__name__.replace('_', ' ')

    def build(self, metadata):
        definitions = super().build(metadata)
        definitions.append(self.asset_definition)
        self.asset_definition['Instantiate Result'] = 'Success'
        return definitions

    @staticmethod
    def _add_asset_metadata(self, attribute_definition, error):
        if _common.present(self.asset_definition, 'Path') and not _common.present(attribute_definition, 'Path'):
            attribute_definition['Path'] = self.asset_definition['Path']

        if _common.present(self.asset_definition, 'Asset') and not _common.present(attribute_definition, 'Asset'):
            attribute_definition['Asset'] = self.asset_definition['Asset']

        if _common.present(self.asset_definition, 'Template') and not _common.present(attribute_definition, 'Template'):
            attribute_definition['Template'] = self.__class__.__name__.replace('_', ' ')

        attribute_definition['Instantiate Result'] = 'Success' if error is None else error

    @classmethod
    def Attribute(cls):
        def attribute_decorator(func):
            def attribute_wrapper(self, metadata):
                func_results = func(self, metadata)

                attribute_definition = dict()

                error = None

                if func_results is None:
                    error = 'None returned by Attribute function'

                def _preserve_originals():
                    for key in ['Name', 'Path']:
                        if _common.present(attribute_definition, key):
                            attribute_definition['Referenced ' + key] = attribute_definition[key]
                            del attribute_definition[key]

                if isinstance(func_results, pd.DataFrame):
                    if len(func_results) == 1:
                        attribute_definition.update(func_results.iloc[0].to_dict())
                        _preserve_originals()
                        attribute_definition['Reference'] = True
                    elif len(func_results) > 1:
                        error = 'Multiple attributes returned by "%s":\n%s' % (func.__name__, func_results)
                    else:
                        error = 'No matching metadata row found for "%s"' % func.__name__

                elif isinstance(func_results, dict):
                    attribute_definition.update(func_results)
                    reference = _common.get(func_results, 'Reference')
                    if reference is not None:
                        if isinstance(reference, pd.DataFrame):
                            if len(reference) == 1:
                                attribute_definition = reference.iloc[0].to_dict()
                                _preserve_originals()
                                attribute_definition['Reference'] = True
                            elif len(reference) > 1:
                                error = 'Multiple attributes returned by "%s":\n%s' % (func.__name__, func_results)
                            else:
                                error = 'No matching metadata found for "%s"' % func.__name__

                if not _common.present(attribute_definition, 'Name'):
                    attribute_definition['Name'] = func.__name__.replace('_', ' ')

                attribute_definition['Asset'] = self.asset_definition['Name']

                Asset._add_asset_metadata(self, attribute_definition, error)

                return attribute_definition

            setattr(attribute_wrapper, 'spy_model', 'attribute')

            return attribute_wrapper

        return attribute_decorator

    @classmethod
    def Component(cls):
        def component_decorator(func):
            def component_wrapper(self, metadata):
                func_results = func(self, metadata)

                component_definitions = list()
                if func_results is None:
                    return component_definitions

                if not isinstance(func_results, list):
                    func_results = [func_results]

                for func_result in func_results:
                    if isinstance(func_result, _AssetBase):
                        _asset_obj = func_result  # type: _AssetBase
                        if not _common.present(_asset_obj.asset_definition, 'Name'):
                            _asset_obj.asset_definition['Name'] = func.__name__.replace('_', ' ')
                        build_results = _asset_obj.build(metadata)
                        component_definitions.extend(build_results)
                    elif isinstance(func_result, dict):
                        component_definition = func_result  # type: dict
                        Asset._add_asset_metadata(self, component_definition, None)
                        component_definitions.append(component_definition)

                return component_definitions

            setattr(component_wrapper, 'spy_model', 'component')

            return component_wrapper

        return component_decorator



import queue
import threading

import pandas as pd
import numpy as np

from seeq.sdk import *

from . import _common
from . import _login


def pull(items, start=None, end=None, grid='15min', header='__auto__', capsules_as='auto', tz_convert=None,
         calculation=None, errors='raise', quiet=False):
    _common.validate_timezone_arg(tz_convert)
    _common.validate_errors_arg(errors)

    if not (isinstance(items, (pd.DataFrame, pd.Series))):
        raise RuntimeError('items argument must be a pandas.DataFrame or pandas.Series')

    if not items.index.is_unique:
        raise RuntimeError("The items DataFrame's index must be unique. Use reset_index(drop=True, inplace=True) "
                           "before passing in to spy.pull().")

    if isinstance(calculation, pd.DataFrame):
        if len(calculation) != 1:
            raise RuntimeError("When applying a calculation across assets, calculation argument must be a one-row "
                               "DataFrame, or a Series. When applying a calculation to a signal/condition/scalar, "
                               'calculation must be a string with a signal variable in it: $signal, $condition or '
                               '$scalar.')

        calculation = calculation.iloc[0]

    if isinstance(items, pd.Series):
        items = pd.DataFrame().append(items)

    if capsules_as not in ['auto', 'capsules', 'signals']:
        raise RuntimeError("capsules_as must be one of 'auto', 'capsules', 'signals'")

    if end is None:
        end = pd.datetime.now()
        if start is not None and start > end:
            end = start + pd.Timedelta(hours=1)

    if start is None:
        start = pd.datetime.now() if end is None else end
        start = start - pd.Timedelta(hours=1)

    pd_start = pd.to_datetime(start)  # type: pd.Timestamp
    pd_end = pd.to_datetime(end)  # type: pd.Timestamp

    status_columns = list()
    if 'ID' in items:
        status_columns.append('ID')
    if 'Path' in items:
        status_columns.append('Path')
    if 'Asset' in items:
        status_columns.append('Asset')
    if 'Name' in items:
        status_columns.append('Name')

    status_df = items[status_columns].copy()
    status_df['Count'] = 0
    status_df['Time'] = 0
    status_df['Result'] = 'Pulling'

    _common.display_status('Pulling data from %s to %s' % (pd_start, pd_end), _common.STATUS_RUNNING, quiet,
                           status_df)

    items_api = ItemsApi(_login.client)
    formulas_api = FormulasApi(_login.client)

    query_df = items  # type: pd.DataFrame
    output_df = pd.DataFrame()
    at_least_one_signal = False
    column_names = dict()
    final_column_names = list()
    for phase in ['signals', 'conditions', 'scalars', 'final']:
        threads = list()
        try:
            pulled_series = queue.Queue()
            status_updates = queue.Queue()

            def _drain_status_updates():
                while True:
                    try:
                        _index, _message, _count, _time = status_updates.get_nowait()
                        status_df.at[_index, 'Result'] = _message
                        status_df.at[_index, 'Count'] = _count
                        status_df.at[_index, 'Time'] = _time
                    except queue.Empty:
                        break

                _common.display_status(
                    'Pulling from <strong>%s</strong> to <strong>%s</strong>' % (pd_start, pd_end),
                    _common.STATUS_RUNNING, quiet, status_df)

            for index, row in query_df.iterrows():
                try:
                    output_df, at_least_one_signal = _process_query_row(
                        _drain_status_updates, at_least_one_signal, calculation, capsules_as,
                        column_names, final_column_names, formulas_api, grid, header, index,
                        items_api, output_df, pd_end, pd_start, phase, pulled_series, row,
                        status_df, status_updates, threads, tz_convert)

                except BaseException as e:
                    if errors == 'raise':
                        raise

                    status_df.at[index, 'Result'] = _common.format_exception(e)

            still_running = True
            while still_running:
                _drain_status_updates()

                still_running = False
                for _, thread in threads:
                    thread.join(0.1)

                    if thread.is_alive():
                        still_running = True
                        break

            _drain_status_updates()

            for item_name, series in list(pulled_series.queue):
                output_df = output_df.join(pd.DataFrame({item_name: series}), how='outer')

        except KeyboardInterrupt:
            for stop_event, _ in threads:
                stop_event.set()

            status_df['Result'] = 'Canceled'
            _common.display_status('Pull canceled', _common.STATUS_CANCELED, quiet, status_df)
            return None

    _common.display_status('Pull successful from <strong>%s</strong> to <strong>%s</strong>' % (pd_start, pd_end),
                           _common.STATUS_SUCCESS, quiet, status_df)

    # We add these columns to the items DataFrame that was passed in so that the caller can inspect the Result field
    # programmatically. In Jupyter, you can see the Result field in the display, so these extra columns are mostly
    # useful in non-Jupyter scenarios. This approach avoids having to return a tuple at all times.
    items['Pull Count'] = status_df['Count']
    items['Pull Time'] = status_df['Time']
    items['Pull Result'] = status_df['Result']

    # Ensures that the order of the columns matches the order in the metadata
    return output_df[final_column_names]


def _process_query_row(_drain_status_updates, at_least_one_signal, calculation, capsules_as, column_names,
                       final_column_names, formulas_api, grid, header, index, items_api, output_df, pd_end, pd_start,
                       phase, pulled_series, row, status_df, status_updates, threads, tz_convert):
    if phase == 'signals' and not _common.present(row, 'ID'):
        status_df.at[index, 'Result'] = 'No "ID" column - skipping'
        return output_df, at_least_one_signal

    item_id, item_name, item_type = _get_item_details(header, items_api, row)
    calculation_to_use = calculation
    if item_type == 'Asset':
        if calculation is None:
            raise RuntimeError('To pull data for an asset, you must provide a "calculate" argument whose '
                               'value is the metadata of a calculation that is based on a single asset.')

        swap_input = SwapInputV1()
        swap_input.swap_in = item_id
        calc_item_id, _, item_type = _get_item_details(header, items_api, calculation)

        item_dependency_output = items_api.get_formula_dependencies(
            id=calc_item_id)  # type: ItemDependencyOutputV1

        unique_assets = set(dep.ancestors[-1].id
                            for dep in item_dependency_output.dependencies
                            if len(dep.ancestors) > 0)

        if len(unique_assets) != 1:
            raise RuntimeError('To pull data for an asset, the "calculate" parameter must be a calculated '
                               'item that involves only one asset.')

        swap_input.swap_out = unique_assets.pop()

        swapped_item = items_api.find_swap(id=calc_item_id, body=[swap_input])  # type: ItemPreviewV1

        item_id = swapped_item.id

        # Don't try to apply a calculation later, we've already done it via our swap activity
        calculation_to_use = None

    if phase == 'signals' and \
            'Signal' not in item_type and 'Condition' not in item_type and 'Scalar' not in item_type:
        status_df.at[index, 'Result'] = 'Not a Signal, Condition or Scalar - skipping'
        return output_df, at_least_one_signal

    if 'Signal' in item_type:
        at_least_one_signal = True

    if phase == 'signals' and 'Signal' in item_type:

        parameters = ['signal=%s' % item_id]
        if calculation_to_use is not None:
            formula = calculation_to_use
        else:
            formula = '$signal'

        if grid:
            formula = 'resample(%s, %s)' % (formula, grid)

        stop_event = threading.Event()
        thread = threading.Thread(target=_pull_signal, args=(formulas_api, formula, parameters, item_id,
                                                             item_name, pulled_series, pd_start, pd_end,
                                                             tz_convert, column_names, stop_event,
                                                             status_updates, index))

        thread.start()
        threads.append((stop_event, thread))

    elif phase == 'conditions' and 'Condition' in item_type:
        if capsules_as == 'capsules' and at_least_one_signal:
            raise RuntimeError('Pull cannot include signals when conditions present and "capsules_as" '
                               'parameter is "capsules"')

        if capsules_as == 'auto':
            capsules_as = 'signals' if at_least_one_signal else 'capsules'

        if capsules_as == 'signals' and not at_least_one_signal:
            if grid is None:
                raise RuntimeError(
                    'Pull cannot include conditions when no signals present with capsules_as='
                    'capsules and grid=None')

            placeholder_item_name = '__placeholder__'
            _pull_signal(formulas_api, '0.toSignal(%s)' % grid, list(), placeholder_item_name,
                         placeholder_item_name, pulled_series, pd_start, pd_end, tz_convert, dict(),
                         threading.Event())

            _, series = pulled_series.get(True, 30)
            output_df[placeholder_item_name] = series

        parameters = ['condition=%s' % item_id]
        if calculation_to_use is not None:
            formula = calculation_to_use
        else:
            formula = '$condition'

        output_df = _pull_condition(capsules_as, formulas_api, formula, parameters, item_id,
                                    item_name, output_df, pd_start, pd_end, tz_convert,
                                    column_names, index, status_df, _drain_status_updates)

    elif phase == 'scalars' and 'Scalar' in item_type:
        parameters = ['scalar=%s' % item_id]
        if calculation_to_use is not None:
            formula = calculation_to_use
        else:
            formula = '$scalar'

        _pull_scalar(formulas_api, formula, parameters, item_id, item_name, output_df, column_names, index,
                     status_df, _drain_status_updates)

    elif phase == 'final':
        if item_id in column_names:
            for column_name in column_names[item_id]:
                if column_name not in final_column_names:
                    final_column_names.append(column_name)

    return output_df, at_least_one_signal


def _convert_column_timezone(ts_column, tz):
    ts_column = ts_column.tz_localize('UTC')
    return ts_column.tz_convert(tz) if tz else ts_column


def _pull_condition(capsules_as, formulas_api, formula, parameters, item_id, item_name, output_df, pd_start, pd_end,
                    tz, column_names, status_index, status_df, _drain_status_updates):
    try:
        timer = _common.timer_start()
        capsule_count = 0
        current_start = pd_start.value
        offset = 0
        while True:
            formula_run_output, _, http_headers = formulas_api.run_formula_with_http_info(
                formula=formula,
                parameters=parameters,
                start='%d ns' % current_start,
                end='%d ns' % pd_end.value,
                offset=offset,
                limit=_common.DEFAULT_PULL_PAGE_SIZE)  # type: FormulaRunOutputV1

            next_start = current_start
            capsules_output = formula_run_output.capsules  # type: CapsulesOutputV1
            check_for_dupes = True

            columns = dict()
            if capsules_as == 'signals':
                columns[item_name] = pd.Series()
                starting_capsule_index = 0
                for _index, _ in output_df.iterrows():
                    for i in range(starting_capsule_index, len(capsules_output.capsules)):
                        capsule = capsules_output.capsules[i]  # type: CapsuleV1
                        capsule_start = _index if pd.isna(capsule.start) else _common.convert_to_timestamp(
                            capsule.start, tz)
                        capsule_end = _index if pd.isna(capsule.end) else _common.convert_to_timestamp(capsule.end, tz)
                        if not pd.isna(capsule.start) and _common.convert_to_timestamp(capsule.start, tz) > next_start:
                            next_start = _common.convert_to_timestamp(capsule.start, tz)
                        present = capsule_start <= _index <= capsule_end
                        if capsule_end < _index:
                            starting_capsule_index = i
                        columns[item_name].at[_index] = 1 if present else 0
                        for prop in capsule.properties:  # type: ScalarPropertyV1
                            colname = '%s - %s' % (item_name, prop.name)
                            if colname not in columns:
                                columns[colname] = pd.Series()
                            columns[colname].at[_index] = prop.value if present else np.nan

                column_names[item_id] = list()
                for col, series in columns.items():
                    output_df[col] = series
                    column_names[item_id].append(col)
            else:
                capsule_df_rows = list()

                for capsule in capsules_output.capsules:
                    column_names[item_id] = ['Condition', 'Capsule Start', 'Capsule End']
                    if check_for_dupes and \
                            'Condition' in output_df and \
                            'Capsule Start' in output_df and \
                            'Capsule End' in output_df and \
                            len(output_df.loc[(output_df['Condition'] == item_name) &
                                              (output_df['Capsule Start'] == _common.convert_to_timestamp(capsule.start,
                                                                                                          tz)) &
                                              (output_df['Capsule End'] == _common.convert_to_timestamp(capsule.end,
                                                                                                        tz))]):
                        # This can happen as a result of pagination
                        continue

                    check_for_dupes = False

                    capsule_dict = {
                        'Condition': item_name,
                        'Capsule Start': _common.convert_to_timestamp(capsule.start, tz),
                        'Capsule End': _common.convert_to_timestamp(capsule.end, tz)
                    }

                    for prop in capsule.properties:  # type: ScalarPropertyV1
                        capsule_dict[prop.name] = prop.value
                        column_names[item_id].append(prop.name)

                    capsule_df_rows.append(capsule_dict)

                    if not pd.isna(capsule.start) and capsule.start > next_start:
                        next_start = capsule.start

                output_df = output_df.append(capsule_df_rows)

            # Note that capsule_count here can diverge from the exact count in the output due to pagination
            capsule_count += len(capsules_output.capsules)

            if len(capsules_output.capsules) < _common.DEFAULT_PULL_PAGE_SIZE:
                break

            if next_start == current_start:
                # This can happen if the page is full of capsule that all have the same start time
                offset += _common.DEFAULT_PULL_PAGE_SIZE
            else:
                offset = 0

            current_start = next_start

            status_df.at[status_index, 'Result'] = 'Pulling %s' % _common.convert_to_timestamp(current_start, tz)
            status_df.at[status_index, 'Count'] = capsule_count
            status_df.at[status_index, 'Time'] = _common.timer_elapsed(timer)

            _drain_status_updates()

        status_df.at[status_index, 'Result'] = 'Success'
        status_df.at[status_index, 'Count'] = capsule_count
        status_df.at[status_index, 'Time'] = _common.timer_elapsed(timer)

    except BaseException as e:
        import traceback
        status_df.at[status_index, 'Result'] = str(traceback.format_exc())

    return output_df


def _pull_signal(formulas_api, formula, parameters, item_id, item_name, pulled_series, pd_start, pd_end, tz,
                 column_names, stop_event=None, status_updates=None, index=None):
    try:
        series = pd.Series()
        timer = _common.timer_start()
        current_start = pd_start
        last_key = 0
        while not stop_event.is_set():
            formula_run_output, _, http_headers = formulas_api.run_formula_with_http_info(
                formula=formula,
                parameters=parameters,
                start='%d ns' % current_start.value,
                end='%d ns' % pd_end.value,
                offset=0,
                limit=_common.DEFAULT_PULL_PAGE_SIZE)  # type: FormulaRunOutputV1

            timings = _common.get_timings(http_headers)

            series_samples_output = formula_run_output.samples  # type: SeriesSamplesOutputV1

            time_index = _convert_column_timezone(pd.DatetimeIndex([sample_output.key for sample_output in
                                                                    series_samples_output.samples if
                                                                    sample_output.key > last_key]), tz)

            series = series.append(pd.Series([sample_output.value for sample_output in
                                              series_samples_output.samples if sample_output.key > last_key],
                                             index=time_index))

            if len(series_samples_output.samples) < _common.DEFAULT_PULL_PAGE_SIZE:
                break

            if len(series) > 0:
                last_key = series.index[-1].value

            if time_index[-1].value > current_start.value:
                current_start = time_index[-1]

            if status_updates is not None:
                status_updates.put((index, 'Pulling: %s' % str(current_start),
                                    len(series), _common.timer_elapsed(timer)))

        column_names[item_id] = [item_name]
        if status_updates is not None:
            status_updates.put((index, 'Success', len(series), _common.timer_elapsed(timer)))
        pulled_series.put((item_name, series))

    except BaseException as e:
        if status_updates is not None:
            status_updates.put((index, str(e), None, None))


def _pull_scalar(formulas_api, formula, parameters, item_id, item_name, output_df, column_names, status_index,
                 status_df, _drain_status_updates):
    timer = _common.timer_start()

    formula_run_output, _, http_headers = formulas_api.run_formula_with_http_info(
        formula=formula,
        parameters=parameters)  # type: FormulaRunOutputV1

    if len(output_df.index) == 0:
        output_df.at[0, item_name] = formula_run_output.scalar.value
    else:
        output_df[item_name] = formula_run_output.scalar.value

    column_names[item_id] = [item_name]

    status_df.at[status_index, 'Result'] = 'Success'
    status_df.at[status_index, 'Count'] = 1
    status_df.at[status_index, 'Time'] = _common.timer_elapsed(timer)


def _get_item_details(header, items_api, row):
    item_id = _common.get(row, 'ID')
    item = None
    if _common.present(row, 'Type'):
        item_type = _common.get(row, 'Type')
    else:
        item = items_api.get_item_and_all_properties(id=item_id)  # type: ItemOutputV1
        item_type = item.type
    if header.upper() == 'ID':
        item_name = item_id
    elif _common.present(row, header):
        item_name = _common.get(row, header)
    else:
        if not item:
            item = items_api.get_item_and_all_properties(id=item_id)  # type: ItemOutputV1

        item_name = ''
        if header == '__auto__' and _common.present(row, 'Path'):
            item_name = _common.get(row, 'Path') + ' >> '
            if _common.present(row, 'Asset'):
                item_name += _common.get(row, 'Asset') + ' >> '

        if header in ['__auto__', 'Name']:
            item_name += item.name
        elif header == 'Description':
            item_name += item.description
        else:
            prop = [p.value for p in item.properties if p.name == header]
            if len(prop) == 0:
                item_name += item_id
            else:
                item_name += prop[0]

    return item_id, item_name, item_type



import os
import re

from . import proputil
from . import system

# Global property keys
SEEQ_INSTALLATION_ID = "seeq_installation_id"
SEEQ_TELEMETRY_ENABLED = "seeq_telemetry_enabled"
SEEQ_SERVICE_MODE = "seeq_service_mode"
SEEQ_ADMIN_CONTACT_NAME = "seeq_admin_contact_name"
SEEQ_ADMIN_CONTACT_EMAIL = "seeq_admin_contact_email"
SEEQ_DATA_FOLDER = "seeq_data_folder"
SEEQ_BACKUP_FOLDER = "seeq_backup_folder"
SEEQ_STARTING_PORT = "seeq_starting_port"
SEEQ_SERVER_HOSTNAME = "seeq_server_hostname"
SEEQ_SERVER_PORT = "seeq_server_port"
SEEQ_SECURE_PORT = "seeq_secure_port"
SEEQ_INCOMING_ALLOWED_HOSTNAME = "seeq_incoming_allowed_hostname"
SEEQ_SERVER_TYPE = "seeq_server_type"

# Ports
APPSERVER_PORT = "appserverPort"
APPSERVER_JMX_PORT = "appserverJMXPort"
APPSERVER_DEBUG_PORT = "appserverDebugPort"
CASSANDRA_CQL_PORT = "cassandraCQLPort"
CASSANDRA_JMX_PORT = "cassandraJMXPort"
CASSANDRA_STORAGE_PORT = "cassandraStoragePort"
CASSANDRA_SSL_PORT = "cassandraSSLPort"
JVM_LINK_JMX_PORT = "JVMLinkJMXPort"
JVM_LINK_DEBUG_PORT = "JVMLinkDebugPort"
WEBSERVER_PRIMARY_PORT = "webserverPrimaryPort"
SUPERVISOR_UI_REST_PORT = "supervisorUIRESTPort"
SUPERVISOR_UI_SINGLE_INSTANCE_PORT = "supervisorUISingleInstancePort"
SUPERVISOR_UI_DEBUG_PORT = "supervisorUIDebugPort"
SUPERVISOR_CORE_REST_PORT = "supervisorCoreRESTPort"
SUPERVISOR_CORE_SINGLE_INSTANCE_PORT = "supervisorCoreSingleInstancePort"
SUPERVISOR_CORE_DEBUG_PORT = "supervisorCoreDebugPort"
IIS_PORT = "iisPort"
IGNITION_DEBUG_PORT = "ignitionDebugPort"
POSTGRES_PORT = "postgresPort"

# Used by cassandra, appserver, and jvm-link; for running in development all the security features turned off
INSECURE_JVM_JMX_OPTIONS = ' '.join([
    '-Dcom.sun.management.jmxremote.authenticate=false',
    '-Dcom.sun.management.jmxremote.ssl=false'
])

ENVIRONMENT_DEV = 'dev'
ENVIRONMENT_PROD = 'prod'

g_seeq_environment = ENVIRONMENT_DEV
g_seeq_global_properties_override = {}
g_seeq_global_folder_override = None


def set_seeq_environment(env):
    global g_seeq_environment
    g_seeq_environment = env


def get_seeq_environment():
    global g_seeq_environment
    return g_seeq_environment


def get_seeq_global_properties_override():
    global g_seeq_global_properties_override
    return g_seeq_global_properties_override


def is_dev():
    return g_seeq_environment == ENVIRONMENT_DEV


def get_global_folder():
    """
    Returns the location of the "global" folder that contains the global properties file, licenses and is the default
    location for the data folder. It is C:\ProgramData\Seeq on Windows, ~/.seeq on UNIX.

    It can be overridden in Pilot commands with the "-g" root argument.

    It can be specified at install time using the "seeq install --global" flag, where it will be written to an
    install.properties file in the image folder.

    :return: Location of the global folder
    """
    global g_seeq_global_folder_override

    global_folder = get_property('seeq_global_folder', get_seeq_install_properties_file())
    if not global_folder:
        if system.is_windows():
            global_folder = os.path.join(os.environ["ProgramData"], 'Seeq')
        else:
            global_folder = os.path.join(system.get_home_dir(), '.seeq')

    if g_seeq_global_folder_override:
        global_folder = g_seeq_global_folder_override

    create_folder_if_necessary_with_correct_permissions(global_folder)

    return global_folder


def override_global_folder(global_folder_override):
    global g_seeq_global_folder_override

    g_seeq_global_folder_override = global_folder_override


def get_seeq_global_properties_file():
    """
    Returns the location of the machine-global properties file that is used for high-level configuration of the
    Seeq product. If not overridden, it is C:\ProgramData\Seeq\global.properties on Windows,
    ~/.seeq/global.properties on UNIX.

    :return: Location of the machine-global properties file
    """
    return os.path.join(get_global_folder(), 'global.properties')


def get_seeq_install_properties_file():
    return os.path.join(get_image_folder(), 'install.properties')


def get_seeq_hostname():
    hostname = 'localhost'
    if get_seeq_global_property(SEEQ_SERVER_HOSTNAME):
        hostname = get_seeq_global_property(SEEQ_SERVER_HOSTNAME)
    return hostname


def get_port(port):
    return get_ports()[port]


def get_ports():
    starting_port = 34210

    properties = get_seeq_global_properties()
    if SEEQ_STARTING_PORT in properties and len(properties[SEEQ_STARTING_PORT].strip()) > 0:
        starting_port = int(properties[SEEQ_STARTING_PORT])

    ports = {
        CASSANDRA_STORAGE_PORT: starting_port,
        CASSANDRA_SSL_PORT: starting_port + 1,
        JVM_LINK_JMX_PORT: starting_port + 2,
        CASSANDRA_JMX_PORT: starting_port + 3,
        CASSANDRA_CQL_PORT: starting_port + 4,
        # starting_port + 5 was previously used for CASSANDRA_THRIFT_PORT and later for ELASTICSEARCH_JMX_PORT
        POSTGRES_PORT: starting_port + 5,
        WEBSERVER_PRIMARY_PORT: starting_port + 6,
        # starting_port + 7 was previously used for webserver fast follow/socket io but is no longer used
        APPSERVER_PORT: starting_port + 8,
        SUPERVISOR_CORE_REST_PORT: starting_port + 9,
        SUPERVISOR_UI_REST_PORT: starting_port + 10,
        SUPERVISOR_CORE_SINGLE_INSTANCE_PORT: starting_port + 11,
        SUPERVISOR_UI_SINGLE_INSTANCE_PORT: starting_port + 12,
        APPSERVER_JMX_PORT: starting_port + 13,
        # starting_port + 14 was previously used for ELASTICSEARCH_REST_PORT but is no longer used
        # starting_port + 15 was previously used for ELASTICSEARCH_NODE_PORT but is no longer used
        IIS_PORT: starting_port + 16,
        APPSERVER_DEBUG_PORT: starting_port + 17,
        JVM_LINK_DEBUG_PORT: starting_port + 18,
        SUPERVISOR_UI_DEBUG_PORT: starting_port + 19,
        SUPERVISOR_CORE_DEBUG_PORT: starting_port + 20}

    if SEEQ_SERVER_PORT in properties and len(properties[SEEQ_SERVER_PORT].strip()) > 0:
        ports[WEBSERVER_PRIMARY_PORT] = int(properties[SEEQ_SERVER_PORT])

    if SEEQ_SECURE_PORT in properties and len(properties[SEEQ_SECURE_PORT].strip()) > 0:
        ports[WEBSERVER_PRIMARY_PORT] = int(properties[SEEQ_SECURE_PORT])

    # These ports are static and unmoving, they are developer-only ports
    ports[IGNITION_DEBUG_PORT] = 34240

    return ports


def override_global_property(prop, value):
    global g_seeq_global_properties_override
    if value is None:
        if prop in g_seeq_global_properties_override:
            del g_seeq_global_properties_override[prop]
    else:
        g_seeq_global_properties_override[prop] = value


def get_seeq_global_properties():
    properties = get_properties(get_seeq_global_properties_file())

    global g_seeq_global_properties_override
    properties.update(g_seeq_global_properties_override)

    return properties


def get_properties(properties_file):
    """
    Retrieves all properties from a properties file.
    :param properties_file: Properties file to use
    :return: A dictionary of property name-values.
    """
    properties = {}

    if not os.path.exists(properties_file):
        return properties

    properties = proputil.get_properties(properties_file)

    return properties


def get_seeq_global_property(property_name):
    global_properties = get_seeq_global_properties()
    return global_properties[property_name] if property_name in global_properties else ''


def get_property(property_name, properties_file):
    """
    Retrieves a property from the machine-global properties file.
    :param property_name: Name of the property to retrieve.
    :param properties_file: Properties file to use
    :return: The value of the property. Will be a blank string ('') if the property is not set or the file does not exist.
    """
    properties = get_properties(properties_file)

    if property_name not in properties:
        return ''

    return properties[property_name]


def set_seeq_global_property(property_name, property_value):
    fix_permissions = False
    if not os.path.exists(get_seeq_global_properties_file()):
        fix_permissions = True

    set_property(property_name, property_value, get_seeq_global_properties_file())

    if fix_permissions:
        system.fix_permissions(get_seeq_global_properties_file())


def set_property(property_name, property_value, properties_file):
    """
    Sets a property in the machine-global properties file.
    :param property_name: Name of the property to set.
    :param properties_file: Properties file to use
    :param property_value: Value of the property.
    """
    properties = get_properties(properties_file)
    properties[property_name] = property_value

    # See comment in propertiesutil.get_properties about why we don't use the pyjavaproperties library.
    lines = []
    for key, value in properties.items():
        lines.append("%s=%s" % (key, value))

    properties_dir = os.path.dirname(properties_file)
    if not os.path.exists(properties_dir):
        os.makedirs(properties_dir)

    lines.sort()

    with open(properties_file, 'w') as f:
        f.writelines(line + '\n' for line in lines)


def check_telemetry_enabled():
    """
    This utility function is called from various sites in the build script, usually when it's likely that AppServer
    is being run. The point is to make sure that the developer is purposefully enabling telemetry, and doesn't
    accidentally leave it on.
    """
    if get_seeq_global_property(SEEQ_TELEMETRY_ENABLED) == '':
        # If it has never been set, then set it to false (on a developer's machine).
        set_seeq_global_property(SEEQ_TELEMETRY_ENABLED, 'false')

    elif get_seeq_global_property(SEEQ_TELEMETRY_ENABLED) == 'true':
        print('************************************************************************************')
        print('')
        print('  HEY!')
        print('')
        print('Telemetry is enabled on your machine in %s' % get_seeq_global_properties_file())
        print('')
        print("Only do this if you're testing the telemetry feature. Otherwise you're polluting")
        print('our telemetry database with developer activity.')
        print('')
        print('************************************************************************************')


def create_folder_if_necessary_with_correct_permissions(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)
        system.fix_permissions(folder)


def get_data_folder():
    seeq_folder = get_global_folder()  # This will ensure the Seeq folder exists no matter what

    data_folder = None
    try:
        data_folder = get_dev_data_folder()
    except BaseException:
        pass

    if not data_folder:
        # Check if the data directory has been configured to a different location
        data_folder = get_seeq_global_property(SEEQ_DATA_FOLDER)
        if not data_folder:
            data_folder = os.path.join(seeq_folder, 'data')

    create_folder_if_necessary_with_correct_permissions(data_folder)

    return data_folder


def set_data_folder(data_folder):
    return set_seeq_global_property(SEEQ_DATA_FOLDER, data_folder)


def get_image_version():
    version_file = os.path.join(get_image_folder(), 'version.txt')
    if not os.path.exists(version_file):
        return None

    with open(version_file, 'r') as f:
        return f.read().strip()


def get_image_version_block():
    image_version = get_image_version()
    match = re.match(r'(\w+)\.(\d+)\.(\d+)\.(\d+)(-v\w+)?(-\w+)?', image_version)
    if not match:
        raise Exception('Unrecognized version string: %s' % image_version)

    version_block = dict()
    version_block['VERSION_PREFIX'] = match.group(1)
    version_block['VERSION_MAJOR'] = match.group(2)
    version_block['VERSION_MINOR'] = match.group(3)
    version_block['VERSION_PATCH'] = match.group(4)
    version_block['VERSION_DATETIME'] = match.group(5)
    version_block['VERSION_SUFFIX'] = match.group(6)

    return version_block


def get_data_version():
    version_file = os.path.join(get_data_folder(), 'version.txt')
    if not os.path.exists(version_file):
        return None

    with open(version_file, 'r') as f:
        return f.read().strip()


def get_dev_data_folder():
    return os.path.join(system.get_repo_root_dir(windows_long_path=False), 'sq-run-data-dir')


def get_image_folder():
    arch = 'x64'
    try:
        arch = system.get_build_architecture()
    except BaseException:
        # We're not executing from within the dev environment
        pass

    production_path = os.path.normpath(os.path.join(os.path.dirname(__file__), '..', '..', '..', '..'))
    development_path = os.path.normpath(
        os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'product', 'image', arch))

    return development_path if os.path.exists(development_path) else production_path


def get_service_exe():
    return os.path.join(get_image_folder(), 'SeeqService.exe')


def get_service_name():
    installation_name = get_property('seeq_installation_name', get_seeq_install_properties_file())
    return 'SeeqService' + ('_' + re.sub(r'\W', '_', installation_name) if installation_name else '')


def get_api_url():
    hostname = get_seeq_hostname()
    protocol = 'http'

    if get_seeq_global_property(SEEQ_SECURE_PORT):
        protocol = 'https'

    return '%s://%s:%d/api' % (protocol, hostname, get_port(WEBSERVER_PRIMARY_PORT))


def get_supervisor_core_properties_path():
    '''
    Retrieve the path to supervisor.properties.
    Will look first in the data folder, then the image folder.
    :return: The path to supervisor.properties or None if it's not found
    '''
    data_path = os.path.join(get_data_folder(), "configuration", "supervisor")
    image_path = os.path.join(get_image_folder(), "supervisor", "image", "configuration")

    filename = 'supervisor.core.properties'
    for path in (data_path, image_path):
        print('Looking for config file "%s" in "%s"' % (filename, path))
        filepath = os.path.join(path, filename)
        if os.path.isfile(filepath):
            print('Using config file "%s' % filepath)
            return filepath

    return None


def get_supervisor_core_memory_multiplier():
    '''
    Get the value of the memory multiplier property from supervisor.core.properties.
    Default to 1.0 if the property cannot be found.
    :return: (float) the memory multiplier value.
    '''
    supervisor_properties_path = get_supervisor_core_properties_path()

    property_string = "memory.multiplier"
    memory_multiplier = None
    if not supervisor_properties_path:
        print("Could not find supervisor.properties.")
    else:
        memory_multiplier = proputil.get_property_from_config(property_string, supervisor_properties_path)

    try:
        memory_multiplier = float(memory_multiplier)
    except TypeError:
        print('Could not find property "%s" in "%s" ' % (property_string, supervisor_properties_path))
        memory_multiplier = 1.0
    except ValueError:
        print('Could not convert %s value "%s" to float' % (property_string, memory_multiplier))
        memory_multiplier = 1.0

    print("Using memory multiplier: %s" % memory_multiplier)

    return float(memory_multiplier)



import datetime
import json
import pytz
import re
import six

from seeq.sdk.rest import ApiException

from IPython.core.display import display, HTML, clear_output

import pandas as pd
import numpy as np

DEFAULT_DATASOURCE_NAME = 'Seeq Data Lab'
DEFAULT_DATASOURCE_CLASS = 'Seeq Data Lab'
DEFAULT_DATASOURCE_ID = 'Seeq Data Lab'
DEFAULT_WORKBOOK_PATH = 'Data Lab >> Data Lab Analysis'
DEFAULT_SEARCH_PAGE_SIZE = 100
DEFAULT_PULL_PAGE_SIZE = 1000000
DEFAULT_PUT_SAMPLES_AND_CAPSULES_BATCH_SIZE = 100000


def present(row, column):
    return (column in row) and \
           (row[column] is not None) and \
           (not isinstance(row[column], float) or not pd.isna(row[column]))


def get(row, column, default=None):
    return row[column] if present(row, column) else default


def get_timings(http_headers):
    output = dict()
    for header, cast in [('Server-Meters', int), ('Server-Timing', float)]:
        server_meters_string = http_headers[header]
        server_meters = server_meters_string.split(',')
        for server_meter_string in server_meters:
            server_meter = server_meter_string.split(';')
            if len(server_meter) < 3:
                continue

            dur_string = cast(server_meter[1].split('=')[1])
            desc_string = server_meter[2].split('=')[1].replace('"', '')

            output[desc_string] = dur_string

    return output


def format_exception(e):
    if isinstance(e, ApiException):
        content = ''
        if e.reason and len(e.reason.strip()) > 0:
            if len(content) > 0:
                content += ' - '
            content += e.reason

        if e.body:
            try:
                body = json.loads(e.body)
                if len(content) > 0:
                    content += ' - '
                content += body['statusMessage']
            except BaseException:
                pass

        return '(%d) %s' % (e.status, content)

    else:
        return str(e)


GUID_REGEX = r'[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12}'


def is_guid(s):
    return re.match(GUID_REGEX, s)


def validate_timezone_arg(tz):
    if tz is not None:
        try:
            pd.to_datetime('2019-01-01T00:00:00.000Z').tz_convert(tz)

        except pytz.exceptions.UnknownTimeZoneError:
            raise RuntimeError('Unknown timezone "%s". Acceptable timezones:\n%s' % (tz, '\n'.join(pytz.all_timezones)))


def validate_errors_arg(errors):
    if errors not in ['raise', 'catalog']:
        raise RuntimeError("errors argument must be either 'raise' or 'catalog'")


def none_to_nan(v):
    return np.nan if v is None else v


def ensure_unicode(o):
    if isinstance(o, six.binary_type):
        return six.text_type(o, 'utf-8', errors='replace')
    else:
        return o


def timer_start():
    return datetime.datetime.now()


def timer_elapsed(timer):
    return datetime.datetime.now() - timer


def convert_to_timestamp(unix_timestamp_in_ns, tz):
    return convert_timestamp_timezone(none_to_nan(pd.Timestamp(unix_timestamp_in_ns)), tz)


def convert_timestamp_timezone(timestamp, tz):
    if pd.isna(timestamp):
        return timestamp

    timestamp = timestamp.tz_localize('UTC')
    return timestamp.tz_convert(tz) if tz else timestamp


def constuct_metrics_dataframe(d):
    return pd.DataFrame(d).transpose()


STATUS_RUNNING = 0
STATUS_SUCCESS = 1
STATUS_FAILURE = 2
STATUS_CANCELED = 3


def display_supports_html():
    # noinspection PyBroadException
    try:
        # noinspection PyUnresolvedReferences
        ipy_str = str(type(get_ipython()))
        if 'zmqshell' in ipy_str:
            return True
        if 'terminal' in ipy_str:
            return False

    except:
        return False


_last_display_status_message = None


def display_status(message, status, quiet, metrics=None):
    if quiet:
        return

    if not display_supports_html():
        global _last_display_status_message
        if message != _last_display_status_message:
            text = re.sub(r'</?[^>]+>', '', message)
            display(text)
            _last_display_status_message = message
        return

    ipython_clear_output(wait=True)

    if status == STATUS_RUNNING:
        color = '#EEEEFF'
    elif status == STATUS_SUCCESS:
        color = '#EEFFEE'
    else:
        color = '#FFEEEE'

    style = 'background-color: %s;' % color
    html = '<div style="%s">%s</div>' % (
        style + 'text-align: left;', message.replace('\n', '<br>'))

    if metrics is not None:
        if isinstance(metrics, pd.DataFrame):
            df = metrics
        else:
            df = constuct_metrics_dataframe(metrics)

        html += '<table>'
        html += '<tr><td style="%s"></td>' % style

        for col in df.columns:
            html += '<td style="%s">%s</td>' % (style, col)

        html += '</tr>'

        for index, row in df.iterrows():
            html += '<tr style="%s">' % style
            html += '<td>%s</td>' % index
            for cell in row:
                if isinstance(cell, datetime.timedelta):
                    hours, remainder = divmod(cell.seconds, 3600)
                    minutes, seconds = divmod(remainder, 60)
                    html += '<td>{:02}:{:02}:{:02}.{:02}</td>'.format(int(hours), int(minutes), int(seconds),
                                                                      int((cell.microseconds + 5000) / 10000))
                else:
                    html += '<td>%s</td>' % cell
            html += '</tr>'

        html += '</table>'

    display(HTML(html))


def ipython_clear_output(wait=False):
    clear_output(wait)


def ipython_display(*objs, include=None, exclude=None, metadata=None, transient=None, display_id=None, **kwargs):
    display(*objs, include=include, exclude=exclude, metadata=metadata, transient=transient,
            display_id=display_id, **kwargs)



import pytest

import pandas as pd
import numpy as np

from seeq import spy
from seeq.sdk.rest import ApiException

from . import test_common


def setup_module():
    test_common.login()


@pytest.mark.system
def test_pull_signal_with_grid():
    search_results = spy.search({
        "Path": "Example >> Cooling Tower 1 >> Area A"
    })

    search_results = search_results.loc[
        search_results['Name'].isin(['Compressor Power', 'Compressor Stage'])]

    df = spy.pull(search_results, start='2019-01-01', end='2019-03-07', grid='5min', header='Name',
                  tz_convert='US/Central')

    # We assert an exact value here to draw attention to any changes. The only reason this number should change is if
    # the Example Data changes in some way or there is a (possibly unexpected) change to Spy.
    assert len(df) == 18721

    # Note that the canonical timezone for US/Central appears to be CST
    assert df.index[0].tzname() == 'CST'

    assert isinstance(df.iloc[0]['Compressor Power'], np.float64)
    assert isinstance(df.iloc[0]['Compressor Stage'], np.str)


@pytest.mark.system
def test_pull_signal_no_grid():
    # This test ensures that time series with non-matching timestamps are returned
    # in a DataFrame with index entries properly interleaved and NaNs where one
    # series has a value and one doesn't.

    data1_df = spy.push(pd.DataFrame({'NoGridData1': [1, 2, 3]},
                                     index=[
                                         pd.to_datetime('2019-01-01T00:00:00.000Z'),
                                         pd.to_datetime('2019-01-01T01:00:00.000Z'),
                                         pd.to_datetime('2019-01-01T02:00:00.000Z'),
                                     ]))

    data2_df = spy.push(pd.DataFrame({'NoGridData2': [10, 20, 30]},
                                     index=[
                                         pd.to_datetime('2019-01-01T00:10:00.000Z'),
                                         pd.to_datetime('2019-01-01T01:10:00.000Z'),
                                         pd.to_datetime('2019-01-01T02:10:00.000Z'),
                                     ]))

    data3_df = spy.push(pd.DataFrame({'NoGridData3': [100, 200, 300]},
                                     index=[
                                         pd.to_datetime('2019-01-01T00:20:00.000Z'),
                                         pd.to_datetime('2019-01-01T01:20:00.000Z'),
                                         pd.to_datetime('2019-01-01T02:20:00.000Z'),
                                     ]))

    all_df = data1_df.append(data2_df).append(data3_df)

    pull_df = spy.pull(all_df, start='2018-12-01', end='2019-12-01', grid=None)

    expected_df = pd.DataFrame({
        'NoGridData1': [1, np.nan, np.nan, 2, np.nan, np.nan, 3, np.nan, np.nan],
        'NoGridData2': [np.nan, 10, np.nan, np.nan, 20, np.nan, np.nan, 30, np.nan],
        'NoGridData3': [np.nan, np.nan, 100, np.nan, np.nan, 200, np.nan, np.nan, 300]
    }, index=[
        pd.to_datetime('2019-01-01T00:00:00.000Z'),
        pd.to_datetime('2019-01-01T00:10:00.000Z'),
        pd.to_datetime('2019-01-01T00:20:00.000Z'),
        pd.to_datetime('2019-01-01T01:00:00.000Z'),
        pd.to_datetime('2019-01-01T01:10:00.000Z'),
        pd.to_datetime('2019-01-01T01:20:00.000Z'),
        pd.to_datetime('2019-01-01T02:00:00.000Z'),
        pd.to_datetime('2019-01-01T02:10:00.000Z'),
        pd.to_datetime('2019-01-01T02:20:00.000Z')
    ])

    assert pull_df.equals(expected_df)


@pytest.mark.system
def test_bad_timezone():
    with pytest.raises(RuntimeError):
        spy.pull(pd.DataFrame(), tz_convert='CDT')


@pytest.mark.system
def test_pull_condition_as_capsules():
    search_results = spy.search({
        'Name': 'Area A_Temperature'
    })

    push_result = spy.push(metadata=pd.DataFrame([{
        'Type': 'Condition',
        'Name': 'Hot',
        'Formula': '$a.validValues().valueSearch(isGreaterThan(80))',
        'Formula Parameters': {
            '$a': search_results.iloc[0]
        }
    }]))

    spy.pull(push_result.iloc[0], start='2019-01-01T00:00:00.000Z', end='2019-06-01T00:00:00.000Z')


@pytest.mark.system
def test_pull_condition_as_signal():
    search_results = spy.search({
        'Name': 'Area A_Temperature'
    })

    push_result = spy.push(metadata=pd.DataFrame([{
        'Type': 'Condition',
        'Name': 'Hot',
        'Formula': '$a.validValues().valueSearch(isGreaterThan(80))',
        'Formula Parameters': {
            '$a': search_results.iloc[0]
        }
    }]))

    spy.pull(push_result.iloc[0], start='2019-01-01T00:00:00.000Z', end='2019-06-01T00:00:00.000Z',
             capsules_as='signals')


@pytest.mark.system
def test_pull_swapped_condition():
    search_results = spy.search({
        'Name': 'Temperature',
        'Path': 'Example >> Cooling Tower 1 >> Area A'
    })

    push_result = spy.push(metadata=pd.DataFrame([{
        'Type': 'Signal',
        'Name': 'Temperature Minus 5',
        'Formula': '$a - 5',
        'Formula Parameters': {
            '$a': search_results.iloc[0]
        }
    }]))

    push_result = spy.push(metadata=pd.DataFrame([{
        'Type': 'Condition',
        'Name': 'Cold',
        'Formula': '$a.validValues().valueSearch(isLessThan(80))',
        'Formula Parameters': {
            '$a': push_result.iloc[0]
        }
    }]))

    pull_df = spy.search({
        'Type': 'Asset',
        'Path': 'Example >> Cooling Tower 2'
    })

    # There will be an error related to trying to swap in Area F
    with pytest.raises(ApiException):
        spy.pull(pull_df, start='2019-01-01T00:00:00.000Z', end='2019-06-01T00:00:00.000Z',
                 calculation=push_result)

    df = spy.pull(pull_df, start='2019-01-01T00:00:00.000Z', end='2019-06-01T00:00:00.000Z',
                  calculation=push_result, errors='catalog')

    assert len(df) > 800

    conditions = df['Condition'].drop_duplicates().tolist()

    assert conditions == [
        'Example >> Cooling Tower 2 >> Area D',
        'Example >> Cooling Tower 2 >> Area E'
    ]



import pytest

from seeq import spy

from . import test_common


def setup_module():
    test_common.login()


@pytest.mark.system
def test_get_folder():
    results_df = spy.workbooks.get_folder(recursive=True)

    print(results_df)



import re

import pandas as pd

from seeq.sdk import *

from . import _common
from . import _login
from . import _push


def search(query, all_properties=False, workbook=_common.DEFAULT_WORKBOOK_PATH, recursive=True, quiet=False):
    """
    Issues a query to the Seeq Server to retrieve metadata for signals,
    conditions, scalars and assets. This metadata can be used to retrieve
    samples, capsules for a particular time range.
    :param query: A dictionary of property / match-criteria pairs. Match
    criteria uses the same syntax as the Data tab in Seeq.
    :type query: dict, pd.DataFrame
    :param all_properties: True if all item properties should be retrieved.
    This currently makes the search operation much slower as retrieval of
    properties for an item requires a separate call.
    :type all_properties: bool
    :param workbook: A path string (with >> delimiters) to indicate a workbook
    for which to scope the search to. Note that globally scoped items will
    still be returned.
    :type workbook: str, None
    :param recursive: If True, searches that include a Path entry will include
    items at and below the specified location in an asset tree. If False,
    the query dictionary can only contain a Path entry and only items at the
    specified level will be returned.
    :type recursive: bool
    :param quiet: If True, suppresses progress output.
    :type quiet: bool
    :return: A DataFrame with rows for each item found and columns for each
    property.
    :rtype: pd.DataFrame
    """
    items_api = ItemsApi(_login.client)
    trees_api = TreesApi(_login.client)

    if isinstance(query, pd.DataFrame):
        queries = query.to_dict(orient='records')
        comparison = '=='
    else:
        queries = [query]
        comparison = '~='

    metadata = list()
    columns = list()
    warnings = list()

    _common.display_status('Initializing',
                           _common.STATUS_RUNNING,
                           quiet,
                           {
                               'Results': {
                                   'Time': 0,
                                   'Count': 0
                               }
                           })

    timer = _common.timer_start()

    def _add_to_dict(d, _key, val):
        d[_key] = _common.none_to_nan(val)

        # We want the columns to appear in a certain order (the order we added them in) for readability
        if _key not in columns:
            columns.append(_key)

    def _get_warning_string():
        if len(warnings) == 0:
            return ''

        return '<em><br>Warning:<br>%s</em>' + '<br>'.join(warnings)

    for query in queries:
        if 'ID' in query:
            # If ID is specified, short-circuit everything and just get the item directly
            _item = items_api.get_item_and_all_properties(id=query['ID'])  # type: ItemOutputV1
            _prop_dict = dict()
            for prop in _item.properties:  # type: PropertyOutputV1
                _add_to_dict(_prop_dict, prop.name, prop.value)

            # Name doesn't seem to appear in additional properties
            _add_to_dict(_prop_dict, 'Name', _item.name)

            metadata.append(_prop_dict)
            continue

        if not recursive:
            if len(query) != 1 or 'Path' not in query:
                raise RuntimeError(
                    'When recursive=False, query dictionary must contain only a Path entry. Use Pandas operations to '
                    'on the returned DataFrame to pare down results.')

        allowed_properties = ['Type', 'Name', 'Description', 'Path', 'Asset', 'Datasource Class', 'Datasource ID',
                              'Datasource Name', 'Cache Enabled', 'Archived', 'Scoped To']

        for key, value in query.items():
            if key not in allowed_properties:
                warnings.append('Property "%s" is not an indexed property and will be ignored. Use any of the '
                                'following searchable properties and then filter further using DataFrame '
                                'operations:\n"%s"' % (key, '", "'.join(allowed_properties)))

        item_types = list()
        clauses = list()

        if _common.present(query, 'Type'):
            item_type_specs = list()
            if isinstance(query['Type'], list):
                item_type_specs.extend(query['Type'])
            else:
                item_type_specs.append(query['Type'])

            for item_type_spec in item_type_specs:
                if item_type_spec == 'Signal':
                    item_types.extend(['StoredSignal', 'CalculatedSignal'])
                elif item_type_spec == 'Condition':
                    item_types.extend(['StoredCondition', 'CalculatedCondition'])
                elif item_type_spec == 'Scalar':
                    item_types.extend(['StoredScalar', 'CalculatedScalar'])
                else:
                    item_types.append(item_type_spec)

            del query['Type']

        if _common.present(query, 'Datasource Name'):
            datasource_results = items_api.search_items(filters=['Name == %s' % _common.get(query, 'Datasource Name')],
                                                        types=['Datasource'],
                                                        limit=100000)  # type: ItemSearchPreviewPaginatedListV1

            if len(datasource_results.items) > 1:
                raise RuntimeError('Multiple datasources found that match "%s"' % _common.get(query, 'Datasource Name'))
            elif len(datasource_results.items) == 0:
                raise RuntimeError('No datasource found that matches "%s"' % _common.get(query, 'Datasource Name'))
            else:
                datasource = datasource_results.items[0]  # type: ItemSearchPreviewV1
                query['Datasource ID'] = items_api.get_property(id=datasource.id, property_name='Datasource ID').value
                query['Datasource Class'] = items_api.get_property(id=datasource.id,
                                                                   property_name='Datasource Class').value

            del query['Datasource Name']

        for prop_name in ['Name', 'Description', 'Datasource Class', 'Datasource ID', 'Data ID']:
            if prop_name in query and query[prop_name] is not None:
                clauses.append(prop_name + comparison + query[prop_name])
                del query[prop_name]

        filters = [' && '.join(clauses)] if len(clauses) > 0 else []

        kwargs = {
            'filters': filters,
            'types': item_types,
            'limit': _common.DEFAULT_SEARCH_PAGE_SIZE
        }

        if workbook:
            if _common.is_guid(workbook):
                workbook_id = workbook
            else:
                workbook_id = _push.reify_workbook(workbook, create=False)

            if workbook_id:
                kwargs['scope'] = workbook_id
            elif workbook != _common.DEFAULT_WORKBOOK_PATH:
                raise RuntimeError('Workbook "%s" not found, or is not accessible by you' % workbook)

        if _common.present(query, 'Asset') and not _common.present(query, 'Path'):
            raise RuntimeError('"Path" query parameter must be present when "Asset" parameter present')

        asset_id = None
        path_parts = None
        if _common.present(query, 'Path'):
            path_parts = re.split(r'\s*>>\s*', query['Path'])
            if _common.present(query, 'Asset'):
                path_parts.append(query['Asset'])
                del query['Asset']

            for path_part in path_parts:
                tree_kwargs = dict()
                if 'scope' in kwargs:
                    tree_kwargs['scoped_to'] = kwargs['scope']

                if not asset_id:
                    tree_output = trees_api.get_tree_root_nodes(**tree_kwargs)  # type: AssetTreeOutputV1
                else:
                    tree_kwargs['id'] = asset_id
                    tree_output = trees_api.get_tree(**tree_kwargs)  # type: AssetTreeOutputV1

                found = False
                for child in tree_output.children:  # type: TreeItemOutputV1
                    if child.name == path_part:
                        found = True
                        asset_id = child.id
                        break

                if not found:
                    raise RuntimeError('Could not found asset "%s"' % query['Path'])

            del query['Path']

        if asset_id:
            kwargs['asset'] = asset_id

        if 'Scoped To' in query and query['Scoped To'] is not None:
            kwargs['scope'] = query['Scoped To']
            kwargs['filters'].append('@excludeGloballyScoped')

        def _do_search(offset):
            kwargs['offset'] = offset
            _common.display_status(
                'Querying Seeq Server for items' + _get_warning_string(),
                _common.STATUS_RUNNING,
                quiet,
                {
                    'Results': {
                        'Time': _common.timer_elapsed(timer),
                        'Count': len(metadata)
                    }
                })

            if recursive:
                return items_api.search_items(**kwargs)
            else:
                kwargs2 = {
                    'id': kwargs['asset'],
                    'offset': kwargs['offset'],
                    'limit': kwargs['limit']
                }

                if 'scope' in kwargs:
                    kwargs2['scoped_to'] = kwargs['scope']

                return trees_api.get_tree(**kwargs2)

        def _gather_results_item_search(result):
            item_search_preview = result  # type: ItemSearchPreviewV1
            prop_dict = dict()

            _add_to_dict(prop_dict, 'ID', item_search_preview.id)
            if len(item_search_preview.ancestors) > 1:
                _add_to_dict(prop_dict, 'Path', ' >> '.join([a.name for a in item_search_preview.ancestors[0:-1]]))
                _add_to_dict(prop_dict, 'Asset', item_search_preview.ancestors[-1].name)
            elif len(item_search_preview.ancestors) == 1:
                _add_to_dict(prop_dict, 'Asset', item_search_preview.ancestors[0].name)

            _add_to_dict(prop_dict, 'Name', item_search_preview.name)
            _add_to_dict(prop_dict, 'Description', item_search_preview.description)
            _add_to_dict(prop_dict, 'Type', item_search_preview.type)
            uom = item_search_preview.value_unit_of_measure if item_search_preview.value_unit_of_measure \
                else item_search_preview.source_value_unit_of_measure
            _add_to_dict(prop_dict, 'Value Unit Of Measure', uom)
            datasource_item_preview = item_search_preview.datasource  # type: ItemPreviewV1
            _add_to_dict(prop_dict, 'Datasource Name',
                         datasource_item_preview.name if datasource_item_preview else None)
            if all_properties:
                item = items_api.get_item_and_all_properties(id=item_search_preview.id)  # type: ItemOutputV1
                for prop in item.properties:  # type: PropertyOutputV1
                    _add_to_dict(prop_dict, prop.name, prop.value)

            metadata.append(prop_dict)

        def _gather_results_get_tree(result):
            tree_item_output = result  # type: TreeItemOutputV1
            prop_dict = dict()

            _add_to_dict(prop_dict, 'ID', tree_item_output.id)
            if len(path_parts) > 1:
                _add_to_dict(prop_dict, 'Path', ' >> '.join(path_parts[0:-1]))
                _add_to_dict(prop_dict, 'Asset', path_parts[-1])
            elif len(path_parts) == 1:
                _add_to_dict(prop_dict, 'Asset', path_parts[0])

            _add_to_dict(prop_dict, 'Name', tree_item_output.name)
            _add_to_dict(prop_dict, 'Description', tree_item_output.description)
            _add_to_dict(prop_dict, 'Type', tree_item_output.type)
            _add_to_dict(prop_dict, 'Value Unit Of Measure', tree_item_output.value_unit_of_measure)
            if all_properties:
                item = items_api.get_item_and_all_properties(id=tree_item_output.id)  # type: ItemOutputV1
                for prop in item.properties:  # type: PropertyOutputV1
                    _add_to_dict(prop_dict, prop.name, prop.value)

            metadata.append(prop_dict)

        if recursive:
            _iterate_over_output(_do_search, 'items', _gather_results_item_search)
        else:
            _iterate_over_output(_do_search, 'children', _gather_results_get_tree)

    _common.display_status(
        'Query successful' + _get_warning_string(),
        _common.STATUS_SUCCESS,
        quiet,
        {
            'Results': {
                'Time': _common.timer_elapsed(timer),
                'Count': len(metadata)
            }
        })

    return pd.DataFrame(data=metadata, columns=columns)


def _iterate_over_output(output_func, collection_name, action_func):
    offset = 0
    while True:
        output = output_func(offset)

        collection = getattr(output, collection_name)

        for item in collection:
            action_func(item)

        if len(collection) != output.limit:
            break

        offset += output.limit



from types import ModuleType

import pandas as pd

from ._model import _AssetBase


def instantiate(model, metadata):
    return pd.DataFrame(_instantiate(model, metadata))


def _instantiate(model, metadata):
    results = list()

    if issubclass(model, _AssetBase):
        if 'Instance Path' not in metadata or 'Instance Asset' not in metadata:
            raise RuntimeError('"Instance Path", "Instance Asset" are required columns')
        if 'Instance Template' in metadata:
            raise RuntimeError('"Instance Template" not allowed when "model" parameter is Asset/Mixin class '
                               'declaration')
        unique_assets = metadata[['Instance Path', 'Instance Asset']].drop_duplicates().dropna()
    elif isinstance(model, ModuleType):
        if 'Instance Path' not in metadata or 'Instance Asset' not in metadata or 'Instance Template' not in metadata:
            raise RuntimeError('"Instance Path", "Instance Asset", "Instance Template" are required columns')
        unique_assets = metadata[['Instance Path', 'Instance Asset', 'Instance Template']].drop_duplicates().dropna()
    else:
        raise RuntimeError('"model" parameter must be a Python module (with Assets/Mixins) or an Asset/Mixin class '
                           'declaration')

    for index, row in unique_assets.iterrows():
        if isinstance(model, ModuleType):
            template = getattr(model, row['Instance Template'].replace(' ', '_'))
        else:
            template = model

        instance = template({
            'Name': row['Instance Asset'],
            'Asset': row['Instance Asset'],
            'Path': row['Instance Path']
        })

        instance_metadata = metadata[(metadata['Instance Asset'] == row['Instance Asset']) &
                                     (metadata['Instance Path'] == row['Instance Path'])]

        results.extend(instance.build(instance_metadata))

    return results



import logging
import os
import six
import urllib
import urllib3

from seeq.base import gconfig
from seeq.sdk import *
from seeq.sdk.rest import ApiException

from . import _common

from urllib3.connectionpool import MaxRetryError

client = None  # type: ApiClient
user = None  # type: UserOutputV1


def login(username=None, password=None, url=None, auth_provider='Seeq', ignore_ssl_errors=False,
          credentials_file=None, quiet=False):
    # Annoying warnings are printed to stderr if connections fail
    logging.getLogger("requests").setLevel(logging.FATAL)
    logging.getLogger("urllib3").setLevel(logging.FATAL)
    urllib3.disable_warnings()

    if url:
        parsed_url = urllib.parse.urlparse(url)
        gconfig.override_global_property('seeq_server_hostname', parsed_url.hostname)
        if parsed_url.scheme == 'https':
            port = '443'
            if parsed_url.port:
                port = six.text_type(parsed_url.port)
            gconfig.override_global_property('seeq_server_port', '')
            gconfig.override_global_property('seeq_secure_port', port)
        if parsed_url.scheme == 'http':
            port = '80'
            if parsed_url.port:
                port = six.text_type(parsed_url.port)
            gconfig.override_global_property('seeq_server_port', port)
            gconfig.override_global_property('seeq_secure_port', '')

    api_client_url = gconfig.get_api_url()

    cert_file = os.path.join(gconfig.get_data_folder(), 'keys', 'seeq-cert.pem')
    if os.path.exists(cert_file):
        Configuration().cert_file = cert_file
    key_file = os.path.join(gconfig.get_data_folder(), 'keys', 'seeq-key.pem')
    if os.path.exists(key_file):
        Configuration().key_file = key_file

    Configuration().verify_ssl = not ignore_ssl_errors

    global client
    client = ApiClient(api_client_url)

    auth_api = AuthApi(client)
    auth_input = AuthInputV1()

    if credentials_file:
        try:
            with open(credentials_file) as f:
                lines = f.readlines()
        except Exception as e:
            raise RuntimeError('Could not read credentials_file "%s": %s' % (credentials_file, e))

        if len(lines) < 2:
            raise RuntimeError('credentials_file "%s" must have two lines: username then password')

        username = lines[0].strip()
        password = lines[1].strip()

    if not username or not password:
        raise RuntimeError('Both username and password must be supplied')

    auth_input.username = username
    auth_input.password = password

    _common.display_status('Logging in to <strong>%s</strong> as <strong>%s</strong>' % (
        api_client_url, username), _common.STATUS_RUNNING, quiet)

    auth_providers = dict()
    try:
        auth_providers_output = auth_api.get_auth_providers()  # type: AuthProvidersOutputV1
    except MaxRetryError as e:
        raise RuntimeError(
            '"%s" could not be reached. Is the server or network down?\n%s' % (api_client_url, e))

    for datasource_output in auth_providers_output.auth_providers:  # type: DatasourceOutputV1
        auth_providers[datasource_output.name] = datasource_output

    if auth_provider not in auth_providers:
        raise RuntimeError('auth_provider "%s" not recognized. Possible auth_provider(s) for this server: %s' %
                           (auth_provider, ', '.join(auth_providers.keys())))

    datasource_output = auth_providers[auth_provider]
    auth_input.auth_provider_class = datasource_output.datasource_class
    auth_input.auth_provider_id = datasource_output.datasource_id

    try:
        auth_api.login(body=auth_input)
    except ApiException as e:
        if e.status == 401:
            raise RuntimeError(
                '"%s" could not be logged in with supplied credentials, check username and password.' %
                auth_input.username)
        else:
            raise
    except MaxRetryError as e:
        raise RuntimeError(
            '"%s" could not be reached. Is the server or network down?\n%s' % (api_client_url, e))
    except Exception as e:
        raise RuntimeError('Could not connect to Seeq"s API at %s with login "%s".\n%s' % (api_client_url,
                                                                                           auth_input.username, e))

    users_api = UsersApi(client)

    global user
    user = users_api.get_me()  # type: UserOutputV1

    user_string = user.username
    user_profile = ''
    if user.first_name:
        user_profile = user.first_name
    if user.last_name:
        user_profile += ' ' + user.last_name
    if user.is_admin:
        user_profile += ' [Admin]'
    if len(user_profile) > 0:
        user_string += ' (%s)' % user_profile.strip()

    _common.display_status('Logged in to <strong>%s</strong> successfully as <strong>%s</strong>' % (
        api_client_url, user_string), _common.STATUS_SUCCESS, quiet)



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: 0.44.00-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat
from six import iteritems
import re


class ServerStatusOutputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'admin_contact_email': 'str',
        'admin_contact_name': 'str',
        'installation_id': 'str',
        'server_specs': 'list[ServerSpecOutputV1]',
        'status': 'str',
        'telemetry_enabled': 'bool',
        'version': 'str'
    }

    attribute_map = {
        'admin_contact_email': 'adminContactEmail',
        'admin_contact_name': 'adminContactName',
        'installation_id': 'installationId',
        'server_specs': 'serverSpecs',
        'status': 'status',
        'telemetry_enabled': 'telemetryEnabled',
        'version': 'version'
    }

    def __init__(self, admin_contact_email=None, admin_contact_name=None, installation_id=None, server_specs=None, status=None, telemetry_enabled=False, version=None):
        """
        ServerStatusOutputV1 - a model defined in Swagger
        """

        self._admin_contact_email = None
        self._admin_contact_name = None
        self._installation_id = None
        self._server_specs = None
        self._status = None
        self._telemetry_enabled = None
        self._version = None

        if admin_contact_email is not None:
          self.admin_contact_email = admin_contact_email
        if admin_contact_name is not None:
          self.admin_contact_name = admin_contact_name
        if installation_id is not None:
          self.installation_id = installation_id
        if server_specs is not None:
          self.server_specs = server_specs
        if status is not None:
          self.status = status
        if telemetry_enabled is not None:
          self.telemetry_enabled = telemetry_enabled
        if version is not None:
          self.version = version

    @property
    def admin_contact_email(self):
        """
        Gets the admin_contact_email of this ServerStatusOutputV1.
        The admin contact's email, which is controlled on the server

        :return: The admin_contact_email of this ServerStatusOutputV1.
        :rtype: str
        """
        return self._admin_contact_email

    @admin_contact_email.setter
    def admin_contact_email(self, admin_contact_email):
        """
        Sets the admin_contact_email of this ServerStatusOutputV1.
        The admin contact's email, which is controlled on the server

        :param admin_contact_email: The admin_contact_email of this ServerStatusOutputV1.
        :type: str
        """

        self._admin_contact_email = admin_contact_email

    @property
    def admin_contact_name(self):
        """
        Gets the admin_contact_name of this ServerStatusOutputV1.
        The admin contact's name, which is controlled on the server

        :return: The admin_contact_name of this ServerStatusOutputV1.
        :rtype: str
        """
        return self._admin_contact_name

    @admin_contact_name.setter
    def admin_contact_name(self, admin_contact_name):
        """
        Sets the admin_contact_name of this ServerStatusOutputV1.
        The admin contact's name, which is controlled on the server

        :param admin_contact_name: The admin_contact_name of this ServerStatusOutputV1.
        :type: str
        """

        self._admin_contact_name = admin_contact_name

    @property
    def installation_id(self):
        """
        Gets the installation_id of this ServerStatusOutputV1.
        The installation identifier for this machine, which is controlled on the server

        :return: The installation_id of this ServerStatusOutputV1.
        :rtype: str
        """
        return self._installation_id

    @installation_id.setter
    def installation_id(self, installation_id):
        """
        Sets the installation_id of this ServerStatusOutputV1.
        The installation identifier for this machine, which is controlled on the server

        :param installation_id: The installation_id of this ServerStatusOutputV1.
        :type: str
        """

        self._installation_id = installation_id

    @property
    def server_specs(self):
        """
        Gets the server_specs of this ServerStatusOutputV1.
        Information about the specs of the server

        :return: The server_specs of this ServerStatusOutputV1.
        :rtype: list[ServerSpecOutputV1]
        """
        return self._server_specs

    @server_specs.setter
    def server_specs(self, server_specs):
        """
        Sets the server_specs of this ServerStatusOutputV1.
        Information about the specs of the server

        :param server_specs: The server_specs of this ServerStatusOutputV1.
        :type: list[ServerSpecOutputV1]
        """

        self._server_specs = server_specs

    @property
    def status(self):
        """
        Gets the status of this ServerStatusOutputV1.
        The server readiness status

        :return: The status of this ServerStatusOutputV1.
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """
        Sets the status of this ServerStatusOutputV1.
        The server readiness status

        :param status: The status of this ServerStatusOutputV1.
        :type: str
        """

        self._status = status

    @property
    def telemetry_enabled(self):
        """
        Gets the telemetry_enabled of this ServerStatusOutputV1.
        True if sending telemetry back to Seeq is enabled

        :return: The telemetry_enabled of this ServerStatusOutputV1.
        :rtype: bool
        """
        return self._telemetry_enabled

    @telemetry_enabled.setter
    def telemetry_enabled(self, telemetry_enabled):
        """
        Sets the telemetry_enabled of this ServerStatusOutputV1.
        True if sending telemetry back to Seeq is enabled

        :param telemetry_enabled: The telemetry_enabled of this ServerStatusOutputV1.
        :type: bool
        """

        self._telemetry_enabled = telemetry_enabled

    @property
    def version(self):
        """
        Gets the version of this ServerStatusOutputV1.
        The version of Seeq running on the server

        :return: The version of this ServerStatusOutputV1.
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """
        Sets the version of this ServerStatusOutputV1.
        The version of Seeq running on the server

        :param version: The version of this ServerStatusOutputV1.
        :type: str
        """

        self._version = version

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, ServerStatusOutputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import os

from seeq import spy

from seeq.base import gconfig


def login():
    key_path = os.path.join(gconfig.get_data_folder(), 'keys', 'agent.key')
    credentials = open(key_path, "r").read().splitlines()

    spy.login(credentials[0], credentials[1])



import os
import pytest

import pandas as pd

from seeq import spy

from . import test_common


def setup_module():
    test_common.login()


@pytest.mark.system
def test_push_signal():
    push_df = pd.DataFrame(pd.Series([
        1,
        2,
        3
    ], index=[
        pd.to_datetime('2019-01-01'),
        pd.to_datetime('2019-01-02'),
        pd.to_datetime('2019-01-03')
    ]), columns=['My Signal'])

    spy.push(push_df)


@pytest.mark.system
def test_bad_calculation():
    with pytest.raises(RuntimeError):
        spy.push(metadata=pd.DataFrame([{
            'Type': 'Signal',
            'Name': 'Bad Calc',
            'Formula': 'hey(nothing)'
        }]))


@pytest.mark.system
def test_push_calculated_signal():
    area_a_signals = spy.search({
        'Path': 'Example >> Cooling Tower 1 >> Area A'
    })

    dew_point_calc = spy.push(metadata=pd.DataFrame([{
        'Type': 'Signal',
        'Name': 'Dew Point',
        # From https://iridl.ldeo.columbia.edu/dochelp/QA/Basic/dewpoint.html
        'Formula': "$T - ((100 - $RH.setUnits(''))/5)",
        'Formula Parameters': {
            '$T': area_a_signals[area_a_signals['Name'] == 'Temperature'],
            '$RH': area_a_signals[area_a_signals['Name'] == 'Relative Humidity']
        }
    }]))

    assert len(dew_point_calc) == 1
    assert 'ID' in dew_point_calc


@pytest.mark.system
def test_push_signal_with_metadata():
    witsml_folder = os.path.dirname(__file__)
    witsml_file = '011_02_0.csv'
    witsml_df = pd.read_csv(os.path.join(witsml_folder, witsml_file))
    timestamp_column = witsml_df.columns[0]
    witsml_df = pd.read_csv(os.path.join(witsml_folder, witsml_file), parse_dates=[timestamp_column])
    witsml_df = witsml_df.drop(list(witsml_df.filter(regex='.*Unnamed.*')), axis=1)
    witsml_df = witsml_df.dropna(axis=1, how='all')
    witsml_df = witsml_df.set_index(timestamp_column)

    metadata = pd.DataFrame({'Header': witsml_df.columns.values})
    metadata['Type'] = 'Signal'
    metadata['Tag'] = metadata['Header'].str.extract(r'(.*)\(')
    metadata['Value Unit Of Measure'] = metadata['Header'].str.extract(r'\((.*)\)')
    metadata['File'] = witsml_file
    metadata['Well Number'] = metadata['File'].str.extract(r'(\d+)_\d+_\d+\.csv')
    metadata['Wellbore ID'] = metadata['File'].str.extract(r'\d+_(\d+)_\d+\.csv')

    metadata = metadata.set_index('Header')

    # Without a Name column, we expect the push metadata to fail
    with pytest.raises(RuntimeError):
        spy.push(data=witsml_df, metadata=metadata)

    metadata['Name'] = "Well_" + metadata['Well Number'] + "_" + "Wellbore_" + \
                       metadata['Wellbore ID'] + "_" + metadata['Tag']

    push_results_df = spy.push(data=witsml_df, metadata=metadata)

    search_results_df = spy.search(push_results_df.iloc[0])

    assert len(search_results_df) == 1
    assert search_results_df.iloc[0]['Name'] == metadata.iloc[0]['Name']
    assert 'Push Result' not in search_results_df
    assert 'Push Count' not in search_results_df
    assert 'Push Time' not in search_results_df

    pull_results_df = spy.pull(search_results_df,
                               start='2016-07-25T15:00:00.000-07:00',
                               end='2019-07-25T17:00:00.000-07:00',
                               grid=None)

    assert len(pull_results_df) == 999

    # noinspection PyUnresolvedReferences
    assert (witsml_df.index == pull_results_df.index).all()

    witsml_list = witsml_df['BITDEP(ft)'].tolist()
    pull_list = pull_results_df['Well_011_Wellbore_02_BITDEP'].tolist()
    assert witsml_list == pull_list



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: 0.44.00-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat
from six import iteritems
import re


class ExportItemsV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'autoupdate_time_range': 'bool',
        'capsule_time': 'bool',
        'export_capsules': 'bool',
        'export_name': 'str',
        'format': 'str',
        'grid_enabled': 'bool',
        'grid_size': 'str',
        'items': 'list[ExportItemV1]',
        'original_timestamps_enabled': 'bool',
        'statistics_enabled': 'bool',
        'swap_in': 'str',
        'swap_out': 'str'
    }

    attribute_map = {
        'autoupdate_time_range': 'autoupdateTimeRange',
        'capsule_time': 'capsuleTime',
        'export_capsules': 'exportCapsules',
        'export_name': 'exportName',
        'format': 'format',
        'grid_enabled': 'gridEnabled',
        'grid_size': 'gridSize',
        'items': 'items',
        'original_timestamps_enabled': 'originalTimestampsEnabled',
        'statistics_enabled': 'statisticsEnabled',
        'swap_in': 'swapIn',
        'swap_out': 'swapOut'
    }

    def __init__(self, autoupdate_time_range=False, capsule_time=False, export_capsules=False, export_name=None, format=None, grid_enabled=True, grid_size=None, items=None, original_timestamps_enabled=False, statistics_enabled=True, swap_in=None, swap_out=None):
        """
        ExportItemsV1 - a model defined in Swagger
        """

        self._autoupdate_time_range = None
        self._capsule_time = None
        self._export_capsules = None
        self._export_name = None
        self._format = None
        self._grid_enabled = None
        self._grid_size = None
        self._items = None
        self._original_timestamps_enabled = None
        self._statistics_enabled = None
        self._swap_in = None
        self._swap_out = None

        if autoupdate_time_range is not None:
          self.autoupdate_time_range = autoupdate_time_range
        if capsule_time is not None:
          self.capsule_time = capsule_time
        if export_capsules is not None:
          self.export_capsules = export_capsules
        if export_name is not None:
          self.export_name = export_name
        if format is not None:
          self.format = format
        if grid_enabled is not None:
          self.grid_enabled = grid_enabled
        if grid_size is not None:
          self.grid_size = grid_size
        if items is not None:
          self.items = items
        if original_timestamps_enabled is not None:
          self.original_timestamps_enabled = original_timestamps_enabled
        if statistics_enabled is not None:
          self.statistics_enabled = statistics_enabled
        if swap_in is not None:
          self.swap_in = swap_in
        if swap_out is not None:
          self.swap_out = swap_out

    @property
    def autoupdate_time_range(self):
        """
        Gets the autoupdate_time_range of this ExportItemsV1.
        Boolean indicating if the time range for export should be updated to 'now' when export is started.

        :return: The autoupdate_time_range of this ExportItemsV1.
        :rtype: bool
        """
        return self._autoupdate_time_range

    @autoupdate_time_range.setter
    def autoupdate_time_range(self, autoupdate_time_range):
        """
        Sets the autoupdate_time_range of this ExportItemsV1.
        Boolean indicating if the time range for export should be updated to 'now' when export is started.

        :param autoupdate_time_range: The autoupdate_time_range of this ExportItemsV1.
        :type: bool
        """

        self._autoupdate_time_range = autoupdate_time_range

    @property
    def capsule_time(self):
        """
        Gets the capsule_time of this ExportItemsV1.
        True if capsule time is displayed, false otherwise.

        :return: The capsule_time of this ExportItemsV1.
        :rtype: bool
        """
        return self._capsule_time

    @capsule_time.setter
    def capsule_time(self, capsule_time):
        """
        Sets the capsule_time of this ExportItemsV1.
        True if capsule time is displayed, false otherwise.

        :param capsule_time: The capsule_time of this ExportItemsV1.
        :type: bool
        """

        self._capsule_time = capsule_time

    @property
    def export_capsules(self):
        """
        Gets the export_capsules of this ExportItemsV1.
        True if the capsule table should be exported, false otherwise.

        :return: The export_capsules of this ExportItemsV1.
        :rtype: bool
        """
        return self._export_capsules

    @export_capsules.setter
    def export_capsules(self, export_capsules):
        """
        Sets the export_capsules of this ExportItemsV1.
        True if the capsule table should be exported, false otherwise.

        :param export_capsules: The export_capsules of this ExportItemsV1.
        :type: bool
        """

        self._export_capsules = export_capsules

    @property
    def export_name(self):
        """
        Gets the export_name of this ExportItemsV1.
        The desired name for the export. This name is used to create an OData endpoint for accessing the data being exported.

        :return: The export_name of this ExportItemsV1.
        :rtype: str
        """
        return self._export_name

    @export_name.setter
    def export_name(self, export_name):
        """
        Sets the export_name of this ExportItemsV1.
        The desired name for the export. This name is used to create an OData endpoint for accessing the data being exported.

        :param export_name: The export_name of this ExportItemsV1.
        :type: str
        """
        if export_name is None:
            raise ValueError("Invalid value for `export_name`, must not be `None`")

        self._export_name = export_name

    @property
    def format(self):
        """
        Gets the format of this ExportItemsV1.
        The desired export output format. Currently only 'xlsx' and 'odata' are supported.

        :return: The format of this ExportItemsV1.
        :rtype: str
        """
        return self._format

    @format.setter
    def format(self, format):
        """
        Sets the format of this ExportItemsV1.
        The desired export output format. Currently only 'xlsx' and 'odata' are supported.

        :param format: The format of this ExportItemsV1.
        :type: str
        """
        if format is None:
            raise ValueError("Invalid value for `format`, must not be `None`")

        self._format = format

    @property
    def grid_enabled(self):
        """
        Gets the grid_enabled of this ExportItemsV1.
        Boolean indicating if the data grid should be included.

        :return: The grid_enabled of this ExportItemsV1.
        :rtype: bool
        """
        return self._grid_enabled

    @grid_enabled.setter
    def grid_enabled(self, grid_enabled):
        """
        Sets the grid_enabled of this ExportItemsV1.
        Boolean indicating if the data grid should be included.

        :param grid_enabled: The grid_enabled of this ExportItemsV1.
        :type: bool
        """

        self._grid_enabled = grid_enabled

    @property
    def grid_size(self):
        """
        Gets the grid_size of this ExportItemsV1.
        The desired sample period for the export. An automatic grid size is used when gridSize is set to 'false'

        :return: The grid_size of this ExportItemsV1.
        :rtype: str
        """
        return self._grid_size

    @grid_size.setter
    def grid_size(self, grid_size):
        """
        Sets the grid_size of this ExportItemsV1.
        The desired sample period for the export. An automatic grid size is used when gridSize is set to 'false'

        :param grid_size: The grid_size of this ExportItemsV1.
        :type: str
        """

        self._grid_size = grid_size

    @property
    def items(self):
        """
        Gets the items of this ExportItemsV1.
        A list of Series and Capsules to be exported

        :return: The items of this ExportItemsV1.
        :rtype: list[ExportItemV1]
        """
        return self._items

    @items.setter
    def items(self, items):
        """
        Sets the items of this ExportItemsV1.
        A list of Series and Capsules to be exported

        :param items: The items of this ExportItemsV1.
        :type: list[ExportItemV1]
        """
        if items is None:
            raise ValueError("Invalid value for `items`, must not be `None`")

        self._items = items

    @property
    def original_timestamps_enabled(self):
        """
        Gets the original_timestamps_enabled of this ExportItemsV1.
        If the original sample period should be used this value will be true. A manual or automatic grid size may be used when false.

        :return: The original_timestamps_enabled of this ExportItemsV1.
        :rtype: bool
        """
        return self._original_timestamps_enabled

    @original_timestamps_enabled.setter
    def original_timestamps_enabled(self, original_timestamps_enabled):
        """
        Sets the original_timestamps_enabled of this ExportItemsV1.
        If the original sample period should be used this value will be true. A manual or automatic grid size may be used when false.

        :param original_timestamps_enabled: The original_timestamps_enabled of this ExportItemsV1.
        :type: bool
        """

        self._original_timestamps_enabled = original_timestamps_enabled

    @property
    def statistics_enabled(self):
        """
        Gets the statistics_enabled of this ExportItemsV1.
        Boolean indicating if the summary statistics should be included. This flag only applies to Excel exports.

        :return: The statistics_enabled of this ExportItemsV1.
        :rtype: bool
        """
        return self._statistics_enabled

    @statistics_enabled.setter
    def statistics_enabled(self, statistics_enabled):
        """
        Sets the statistics_enabled of this ExportItemsV1.
        Boolean indicating if the summary statistics should be included. This flag only applies to Excel exports.

        :param statistics_enabled: The statistics_enabled of this ExportItemsV1.
        :type: bool
        """

        self._statistics_enabled = statistics_enabled

    @property
    def swap_in(self):
        """
        Gets the swap_in of this ExportItemsV1.
        The ID of an asset to swap in. Any parameters in the formula that are named the same in both the swapIn and swapOut assets will be swapped.

        :return: The swap_in of this ExportItemsV1.
        :rtype: str
        """
        return self._swap_in

    @swap_in.setter
    def swap_in(self, swap_in):
        """
        Sets the swap_in of this ExportItemsV1.
        The ID of an asset to swap in. Any parameters in the formula that are named the same in both the swapIn and swapOut assets will be swapped.

        :param swap_in: The swap_in of this ExportItemsV1.
        :type: str
        """

        self._swap_in = swap_in

    @property
    def swap_out(self):
        """
        Gets the swap_out of this ExportItemsV1.
        The ID of an asset to swap out. Any parameters in the formula that are named the same in both the swapIn and swapOut assets will be swapped.

        :return: The swap_out of this ExportItemsV1.
        :rtype: str
        """
        return self._swap_out

    @swap_out.setter
    def swap_out(self, swap_out):
        """
        Sets the swap_out of this ExportItemsV1.
        The ID of an asset to swap out. Any parameters in the formula that are named the same in both the swapIn and swapOut assets will be swapped.

        :param swap_out: The swap_out of this ExportItemsV1.
        :type: str
        """

        self._swap_out = swap_out

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, ExportItemsV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="seeq",
    version="0.0.35",
    author="Seeq Corporation",
    author_email="support@seeq.com",
    description="The Seeq SDK for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.seeq.com",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        'certifi',
        'ipython',
        'matplotlib',
        'numpy',
        'pandas',
        'six',
        'urllib3'
    ],
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
)



import os

from distutils import dir_util


def copy(directory=None):
    if not directory:
        directory = os.path.join(os.getcwd(), 'Spy Documentation')

    if os.path.exists(directory):
        dir_util.remove_tree(directory)

    dir_util.copy_tree(os.path.join(os.path.dirname(__file__), 'Documentation'), directory)

    print('Copied Spy library documentation to "%s"' % directory)
